package com.css.simulation.resource.common.util;

import api.common.pojo.constants.DictConstants;
import api.common.pojo.param.RedisParameter;
import api.common.util.ObjectUtil;
import com.css.simulation.resource.common.controller.RedisController;
import com.css.simulation.resource.common.util.processbar.ProgressInputStream;
import io.minio.*;
import io.minio.errors.*;
import io.minio.http.Method;
import io.minio.messages.Item;
import jdk.nashorn.internal.runtime.regexp.joni.exception.InternalException;
import org.springframework.web.multipart.MultipartFile;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.List;

public class MinioUtil {

    /**
     * 判断 bucket 是否存在
     */
    public static boolean isBucketExist(MinioClient minioClient, String bucketName) throws InsufficientDataException, ErrorResponseException, IOException, NoSuchAlgorithmException, InvalidKeyException, InvalidResponseException, XmlParserException, InternalException, io.minio.errors.ServerException, io.minio.errors.InternalException {
        return minioClient.bucketExists(BucketExistsArgs.builder().bucket(bucketName).build());
    }

    /**
     * 创建 bucket
     */
    public static void createBucket(MinioClient minioClient, String bucketName) throws InsufficientDataException, ErrorResponseException, IOException, NoSuchAlgorithmException, InvalidKeyException, InvalidResponseException, XmlParserException, InternalException, io.minio.errors.ServerException, io.minio.errors.InternalException {
        if (!isBucketExist(minioClient, bucketName)) {
            minioClient.makeBucket(MakeBucketArgs.builder().bucket(bucketName).build());
        }
    }

    /**
     * 获取文件列表
     *
     * @param minioClient minio 客户端
     * @param bucket      bucket 名称
     * @param prefix      目录名
     * @return 文件路径列表
     */
    public static List<String> listObjects(MinioClient minioClient, String bucket, String prefix) throws ServerException, InsufficientDataException, ErrorResponseException, IOException, NoSuchAlgorithmException, InvalidKeyException, InvalidResponseException, XmlParserException, io.minio.errors.InternalException {
        Iterable<Result<Item>> objects = minioClient.listObjects(
                ListObjectsArgs.builder()
                        .bucket(bucket)
                        .recursive(true)
                        .prefix(prefix)
                        .build()
        );
        List<String> result = new ArrayList<>();
        for (Result<Item> next : objects) {
            Item item = next.get();
            String objectName = item.objectName();
            result.add(objectName);
        }

        return result;
    }



    /**
     * 获取预览路径
     *
     * @return 预览路径
     */
    public static String getPreviewUrl(MinioClient minioClient, Method method, String bucket, String object) throws io.minio.errors.ServerException, InsufficientDataException, ErrorResponseException, IOException, NoSuchAlgorithmException, InvalidKeyException, InvalidResponseException, XmlParserException, io.minio.errors.InternalException {
        return minioClient.getPresignedObjectUrl(GetPresignedObjectUrlArgs.builder()
                .method(method)
                .bucket(bucket)
                .object(object)
                .expiry(3600)
                .build());
    }

    /**
     * 通过文件路径上传文件上传文件
     */
    public static void uploadFromFile(
            MinioClient minioClient,
            String sourceFilePath,
            String bucketName,
            String objectName
    ) throws IOException, InsufficientDataException, ErrorResponseException, NoSuchAlgorithmException, InvalidKeyException, InvalidResponseException, XmlParserException, InternalException, io.minio.errors.ServerException, io.minio.errors.InternalException {
        minioClient.uploadObject(UploadObjectArgs.builder()
                .filename(sourceFilePath)
                .bucket(bucketName)
                .object(objectName)
                .build());
    }

    /**
     * 通过文件路径上传文件上传文件
     *
     * @param partSize 分片最小 5MB
     */
    public static void uploadFromStream(
            MinioClient minioClient,
            InputStream inputStream,
            long objectSize,
            long partSize,
            String bucketName,
            String objectName
    ) throws IOException, InsufficientDataException, ErrorResponseException, NoSuchAlgorithmException, InvalidKeyException, InvalidResponseException, XmlParserException, InternalException, io.minio.errors.ServerException, io.minio.errors.InternalException {
        minioClient.putObject(PutObjectArgs.builder()
                .stream(inputStream, objectSize, partSize)
                .bucket(bucketName)
                .object(objectName)
                .build());
    }

    /**
     * 通过文件路径上传文件上传文件
     */
    public static void uploadFromMultipartFile(
            MinioClient minioClient,
            MultipartFile multipartFile,
            String bucketName,
            String objectName
    ) throws IOException, InsufficientDataException, ErrorResponseException, NoSuchAlgorithmException, InvalidKeyException, InvalidResponseException, XmlParserException, InternalException, io.minio.errors.ServerException, io.minio.errors.InternalException {
        InputStream inputStream = multipartFile.getInputStream();
        long objectSize = multipartFile.getSize();
//        long partSize = 5 * 1024 * 1024L; // 分片最小 5M
        long partSize = -1; // 不分片
        minioClient.putObject(PutObjectArgs.builder()
                .stream(inputStream, objectSize, partSize)
                .bucket(bucketName)
                .object(objectName)
                .build());
    }

    /**
     * 文件上传，保存进度
     */
    public static void uploadProcessBar(
            MinioClient minioClient,
            MultipartFile multipartFile,
            String bucketName,
            String objectName,
            String objectPath,
            RedisController redisController
    ) throws IOException, InsufficientDataException, ErrorResponseException, NoSuchAlgorithmException, InvalidKeyException, InvalidResponseException, XmlParserException, InternalException, io.minio.errors.ServerException, io.minio.errors.InternalException {
        InputStream inputStream = multipartFile.getInputStream();
        long fileSize = multipartFile.getSize();
        //进度计算参数：传输量，进度标尺，文件大小，进度值
        long[] longs = {0L, 0L, fileSize, 0L};
        InputStream pis = new BufferedInputStream(
                new ProgressInputStream(inputStream){
                    @Override
                    public void getStep(int readBytes){
                        RedisParameter redisParameter = getProgress(readBytes,longs);
                        if(ObjectUtil.isNotNull(redisParameter)){
                            redisParameter.setKey(DictConstants.ALGORITHM_KEY + objectPath);//进度缓存的 key
                            redisController.set(redisParameter);
                        }
                    }
                });
        minioClient.putObject(PutObjectArgs.builder()
                .bucket(bucketName)
                .object(objectName)
                .stream(pis, fileSize, -1)
                .build());
        pis.close();

    }

    /**
     * 下载文件
     */
    public static void downloadToFile(
            MinioClient minioClient,
            String bucketName,
            String objectName,
            String targetFilePath
    ) throws IOException, InsufficientDataException, ErrorResponseException, NoSuchAlgorithmException, InvalidKeyException, InvalidResponseException, XmlParserException, InternalException, io.minio.errors.ServerException, io.minio.errors.InternalException {
        minioClient.downloadObject(DownloadObjectArgs.builder()
                .bucket(bucketName)
                .object(objectName)
                .filename(targetFilePath)
                .build());
    }

    /**
     * 下载文件
     */
    public static InputStream downloadToStream(
            MinioClient minioClient,
            String bucketName,
            String objectName
    ) throws IOException, InsufficientDataException, ErrorResponseException, NoSuchAlgorithmException, InvalidKeyException, InvalidResponseException, XmlParserException, InternalException, io.minio.errors.ServerException, io.minio.errors.InternalException {
        return minioClient.getObject(GetObjectArgs.builder()
                .bucket(bucketName)
                .object(objectName)
                .build());
    }

    public static RedisParameter getProgress(int readBytes,long[] longs){
        double progressStep = 0.05;//进度条的步长
        longs[0] += readBytes;//累加
        if(longs[0] >= longs[1]){
            longs[3] += 1;
            double progress = progressStep * longs[3];//进度
            longs[1] = (long) (longs[2] * progress);//标尺增长
            RedisParameter redisParameter = new RedisParameter();
            redisParameter.setValue(String.format("%.2f",progress-progressStep));
            redisParameter.setMinutes(1);
            if(longs[0] == longs[2]){
                redisParameter.setValue("1.00");
            }
            return redisParameter;
        }
        return null;
    }
}
